@@ -18,6 +18,7 @@
 
 #include <config.h>
 
+/* Specification.  */
 #include <unistd.h>
 
 #include "signature.h"
@@ -41,5 +42,14 @@ main (void)
 #endif
   test_getlogin_result (buf, err);
 
+  /* Check that getlogin() does not merely return getenv ("LOGNAME").  */
+  {
+    static char set_LOGNAME[] = "LOGNAME=ygvfibmslhkmvoetbrcegzwydorcke";
+    putenv (set_LOGNAME);
+    buf = getlogin ();
+    ASSERT (!(buf != NULL
+              && strcmp (buf, "ygvfibmslhkmvoetbrcegzwydorcke") == 0));
+  }
+
   return test_exit_status;
 }
