@@ -1,6 +1,6 @@
 /* exclude.c -- exclude file names
 
-   Copyright (C) 1992-1994, 1997, 1999-2007, 2009-2024 Free Software
+   Copyright (C) 1992-1994, 1997, 1999-2007, 2009-2025 Free Software
    Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
@@ -36,6 +36,7 @@
 #include "filename.h"
 #include <fnmatch.h>
 #include "hash.h"
+#include "hashcode-string1.h"
 #if GNULIB_MCEL_PREFER
 # include "mcel.h"
 #else
@@ -239,7 +240,7 @@ string_hasher_ci (void const *data, size_t n_buckets)
 static bool
 string_compare (void const *data1, void const *data2)
 {
-  return strcmp (data1, data2) == 0;
+  return streq (data1, data2);
 }
 
 /* compare two strings for equality, case-insensitive */
@@ -260,7 +261,7 @@ new_exclude_segment (struct exclude *ex, enum exclude_type type, int options)
   switch (type)
     {
     case exclude_pattern:
-      sp->v.pat = (struct exclude_pattern) {0};
+      sp->v.pat = (struct exclude_pattern) { NULL, 0, 0 };
       break;
 
     case exclude_hash:
@@ -658,7 +659,7 @@ add_exclude_file (void (*add_func) (struct exclude *, char const *, int),
                   struct exclude *ex, char const *file_name, int options,
                   char line_end)
 {
-  if (strcmp (file_name, "-") == 0)
+  if (streq (file_name, "-"))
     return add_exclude_fp (call_addfn, ex, stdin, options, line_end, &add_func);
 
   FILE *in = fopen (file_name, "re");
