@@ -1,10 +1,10 @@
 /* userspec.c -- Parse a user and group string.
-   Copyright (C) 1989-1992, 1997-1998, 2000, 2002-2010 Free Software
+   Copyright (C) 1989-1992, 1997-1998, 2000, 2002-2025 Free Software
    Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 3 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -13,7 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* Written by David MacKenzie <djm@gnu.ai.mit.edu>.  */
 
@@ -22,7 +22,6 @@
 /* Specification.  */
 #include "userspec.h"
 
-#include <stdbool.h>
 #include <stdio.h>
 #include <sys/types.h>
 #include <pwd.h>
@@ -44,7 +43,7 @@
 #include "xstrtol.h"
 
 #include "gettext.h"
-#define _(msgid) gettext (msgid)
+#define _(msgid) dgettext (GNULIB_TEXT_DOMAIN, msgid)
 #define N_(msgid) msgid
 
 #ifndef HAVE_ENDGRENT
@@ -79,8 +78,8 @@
    - It's typically faster.
    POSIX says that only '0' through '9' are digits.  Prefer ISDIGIT to
    isdigit unless it's important to use the locale's definition
-   of `digit' even when the host does not conform to POSIX.  */
-#define ISDIGIT(c) ((unsigned int) (c) - '0' <= 9)
+   of "digit" even when the host does not conform to POSIX.  */
+# define ISDIGIT(c) ((unsigned int) (c) - '0' <= 9)
 
 /* Return true if STR represents an unsigned decimal integer.  */
 
@@ -103,10 +102,6 @@ parse_with_separator (char const *spec, char const *separator,
                       uid_t *uid, gid_t *gid,
                       char **username, char **groupname)
 {
-  static const char *E_invalid_user = N_("invalid user");
-  static const char *E_invalid_group = N_("invalid group");
-  static const char *E_bad_spec = N_("invalid spec");
-
   const char *error_msg;
   struct passwd *pwd;
   struct group *grp;
@@ -114,10 +109,13 @@ parse_with_separator (char const *spec, char const *separator,
   char const *g;
   char *gname = NULL;
   uid_t unum = *uid;
-  gid_t gnum = *gid;
+  gid_t gnum = gid ? *gid : -1;
 
   error_msg = NULL;
-  *username = *groupname = NULL;
+  if (username)
+    *username = NULL;
+  if (groupname)
+    *groupname = NULL;
 
   /* Set U and G to nonzero length strings corresponding to user and
      group specifiers or to NULL.  If U is not NULL, it is a newly
@@ -131,10 +129,10 @@ parse_with_separator (char const *spec, char const *separator,
     }
   else
     {
-      size_t ulen = separator - spec;
+      idx_t ulen = separator - spec;
       if (ulen != 0)
         {
-          u = xmemdup (spec, ulen + 1);
+          u = ximemdup (spec, ulen + 1);
           u[ulen] = '\0';
         }
     }
@@ -145,7 +143,7 @@ parse_with_separator (char const *spec, char const *separator,
 
 #ifdef __DJGPP__
   /* Pretend that we are the user U whose group is G.  This makes
-     pwd and grp functions ``know'' about the UID and GID of these.  */
+     pwd and grp functions "know" about the UID and GID of these.  */
   if (u && !is_number (u))
     setenv ("USER", u, 1);
   if (g && !is_number (g))
@@ -158,12 +156,13 @@ parse_with_separator (char const *spec, char const *separator,
       pwd = (*u == '+' ? NULL : getpwnam (u));
       if (pwd == NULL)
         {
+          username = NULL;
           bool use_login_group = (separator != NULL && g == NULL);
           if (use_login_group)
             {
               /* If there is no group,
                  then there may not be a trailing ":", either.  */
-              error_msg = E_bad_spec;
+              error_msg = N_("invalid spec");
             }
           else
             {
@@ -172,7 +171,7 @@ parse_with_separator (char const *spec, char const *separator,
                   && tmp <= MAXUID && (uid_t) tmp != (uid_t) -1)
                 unum = tmp;
               else
-                error_msg = E_invalid_user;
+                error_msg = N_("invalid user");
             }
         }
       else
@@ -199,12 +198,13 @@ parse_with_separator (char const *spec, char const *separator,
       grp = (*g == '+' ? NULL : getgrnam (g));
       if (grp == NULL)
         {
+          groupname = NULL;
           unsigned long int tmp;
           if (xstrtoul (g, NULL, 10, &tmp, "") == LONGINT_OK
               && tmp <= MAXGID && (gid_t) tmp != (gid_t) -1)
             gnum = tmp;
           else
-            error_msg = E_invalid_group;
+            error_msg = N_("invalid group");
         }
       else
         gnum = grp->gr_gid;
@@ -215,46 +215,58 @@ parse_with_separator (char const *spec, char const *separator,
   if (error_msg == NULL)
     {
       *uid = unum;
-      *gid = gnum;
-      *username = u;
-      *groupname = gname;
-      u = NULL;
+      if (gid)
+        *gid = gnum;
+      if (username)
+        {
+          *username = u;
+          u = NULL;
+        }
+      if (groupname)
+        {
+          *groupname = gname;
+          gname = NULL;
+        }
     }
-  else
-    free (gname);
 
   free (u);
-  return _(error_msg);
+  free (gname);
+  return error_msg ? _(error_msg) : NULL;
 }
 
 /* Extract from SPEC, which has the form "[user][:.][group]",
    a USERNAME, UID U, GROUPNAME, and GID G.
+   If the GID parameter is NULL the entire SPEC is treated as a user.
+   If the USERNAME and GROUPNAME parameters are NULL they're ignored.
    Either user or group, or both, must be present.
    If the group is omitted but the separator is given,
    use the given user's login group.
-   If SPEC contains a `:', then use that as the separator, ignoring
-   any `.'s.  If there is no `:', but there is a `.', then first look
+   If SPEC contains a ':', then use that as the separator, ignoring
+   any '.'s.  If there is no ':', but there is a '.', then first look
    up the entire SPEC as a login name.  If that look-up fails, then
-   try again interpreting the `.'  as a separator.
+   try again interpreting the '.'  as a separator.
 
    USERNAME and GROUPNAME will be in newly malloc'd memory.
    Either one might be NULL instead, indicating that it was not
    given and the corresponding numeric ID was left unchanged.
 
-   Return NULL if successful, a static error message string if not.  */
+   Return NULL if successful, a static error message string if not.
+   If PWARN is null, return NULL instead of a warning;
+   otherwise, set *PWARN to true depending on whether returning a warning.  */
 
 char const *
-parse_user_spec (char const *spec, uid_t *uid, gid_t *gid,
-                 char **username, char **groupname)
+parse_user_spec_warn (char const *spec, uid_t *uid, gid_t *gid,
+                      char **username, char **groupname, bool *pwarn)
 {
-  char const *colon = strchr (spec, ':');
+  char const *colon = gid ? strchr (spec, ':') : NULL;
   char const *error_msg =
     parse_with_separator (spec, colon, uid, gid, username, groupname);
+  bool warn = false;
 
-  if (!colon && error_msg)
+  if (gid && !colon && error_msg)
     {
       /* If there's no colon but there is a dot, and if looking up the
-         whole spec failed (i.e., the spec is not a owner name that
+         whole spec failed (i.e., the spec is not an owner name that
          includes a dot), then try again, but interpret the dot as a
          separator.  This is a compatible extension to POSIX, since
          the POSIX-required behavior is always tried first.  */
@@ -262,12 +274,26 @@ parse_user_spec (char const *spec, uid_t *uid, gid_t *gid,
       char const *dot = strchr (spec, '.');
       if (dot
           && ! parse_with_separator (spec, dot, uid, gid, username, groupname))
-        error_msg = NULL;
+        {
+          warn = true;
+          error_msg = pwarn ? N_("warning: '.' should be ':'") : NULL;
+        }
     }
 
+  if (pwarn)
+    *pwarn = warn;
   return error_msg;
 }
 
+/* Like parse_user_spec_warn, but generate only errors; no warnings.  */
+
+char const *
+parse_user_spec (char const *spec, uid_t *uid, gid_t *gid,
+                 char **username, char **groupname)
+{
+  return parse_user_spec_warn (spec, uid, gid, username, groupname, NULL);
+}
+
 #ifdef TEST
 
 # define NULL_CHECK(s) ((s) == NULL ? "(null)" : (s))
