@@ -1,5 +1,5 @@
 /* fold -- wrap each input line to fit in specified width.
-   Copyright (C) 1991-2013 Free Software Foundation, Inc.
+   Copyright (C) 1991-2025 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,21 +12,23 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* Written by David MacKenzie, djm@gnu.ai.mit.edu. */
 
 #include <config.h>
 
+#include <ctype.h>
 #include <stdio.h>
 #include <getopt.h>
 #include <sys/types.h>
 
 #include "system.h"
-#include "error.h"
 #include "fadvise.h"
-#include "quote.h"
-#include "xstrtol.h"
+#include "ioblksize.h"
+#include "mcel.h"
+#include "mbbuf.h"
+#include "xdectoint.h"
 
 #define TAB_WIDTH 8
 
@@ -38,22 +40,31 @@
 /* If nonzero, try to break on whitespace. */
 static bool break_spaces;
 
-/* If nonzero, count bytes, not column positions. */
-static bool count_bytes;
+/* Mode to operate in.  */
+static enum
+  {
+    COUNT_COLUMNS,
+    COUNT_BYTES,
+    COUNT_CHARACTERS
+  } counting_mode = COUNT_COLUMNS;
 
 /* If nonzero, at least one of the files we read was standard input. */
 static bool have_read_stdin;
 
-static char const shortopts[] = "bsw:0::1::2::3::4::5::6::7::8::9::";
+/* Width of last read character.  */
+static int last_character_width = 0;
+
+static char const shortopts[] = "bcsw:0::1::2::3::4::5::6::7::8::9::";
 
 static struct option const longopts[] =
 {
-  {"bytes", no_argument, NULL, 'b'},
-  {"spaces", no_argument, NULL, 's'},
-  {"width", required_argument, NULL, 'w'},
+  {"bytes", no_argument, nullptr, 'b'},
+  {"characters", no_argument, nullptr, 'c'},
+  {"spaces", no_argument, nullptr, 's'},
+  {"width", required_argument, nullptr, 'w'},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
-  {NULL, 0, NULL, 0}
+  {nullptr, 0, nullptr, 0}
 };
 
 void
@@ -68,20 +79,21 @@ Usage: %s [OPTION]... [FILE]...\n\
 "),
               program_name);
       fputs (_("\
-Wrap input lines in each FILE (standard input by default), writing to\n\
-standard output.\n\
+Wrap input lines in each FILE, writing to standard output.\n\
 "), stdout);
 
+      emit_stdin_note ();
       emit_mandatory_arg_note ();
 
       fputs (_("\
   -b, --bytes         count bytes rather than columns\n\
+  -c, --characters    count characters rather than columns\n\
   -s, --spaces        break at spaces\n\
   -w, --width=WIDTH   use WIDTH columns instead of 80\n\
 "), stdout);
       fputs (HELP_OPTION_DESCRIPTION, stdout);
       fputs (VERSION_OPTION_DESCRIPTION, stdout);
-      emit_ancillary_info ();
+      emit_ancillary_info (PROGRAM_NAME);
     }
   exit (status);
 }
@@ -91,27 +103,45 @@ standard output.\n\
    The first column is 0. */
 
 static size_t
-adjust_column (size_t column, char c)
+adjust_column (size_t column, mcel_t g)
 {
-  if (!count_bytes)
+  if (counting_mode != COUNT_BYTES)
     {
-      if (c == '\b')
+      if (g.ch == '\b')
         {
           if (column > 0)
-            column--;
+            column -= last_character_width;
         }
-      else if (c == '\r')
+      else if (g.ch == '\r')
         column = 0;
-      else if (c == '\t')
+      else if (g.ch == '\t')
         column += TAB_WIDTH - column % TAB_WIDTH;
-      else /* if (isprint (c)) */
-        column++;
+      else
+        {
+          if (counting_mode == COUNT_CHARACTERS)
+            last_character_width = 1;
+          else
+            {
+              int width = c32width (g.ch);
+              /* Default to a width of 1 if there is an invalid character.  */
+              last_character_width = width < 0 ? 1 : width;
+            }
+          column += last_character_width;
+        }
     }
   else
-    column++;
+    column += g.len;
   return column;
 }
 
+static void
+write_out (char const *line, size_t line_len, bool newline)
+{
+  if (fwrite (line, sizeof (char), line_len, stdout) != line_len
+      || (newline && putchar ('\n') < 0))
+    write_error ();
+}
+
 /* Fold file FILENAME, or standard input if FILENAME is "-",
    to stdout, with maximum line length WIDTH.
    Return true if successful.  */
@@ -120,14 +150,14 @@ static bool
 fold_file (char const *filename, size_t width)
 {
   FILE *istream;
-  int c;
   size_t column = 0;		/* Screen column where next char will go. */
-  size_t offset_out = 0;	/* Index in 'line_out' for next char. */
-  static char *line_out = NULL;
-  static size_t allocated_out = 0;
+  idx_t offset_out = 0;		/* Index in 'line_out' for next char. */
+  static char line_out[IO_BUFSIZE];
+  static char line_in[IO_BUFSIZE];
+  mbbuf_t mbbuf;
   int saved_errno;
 
-  if (STREQ (filename, "-"))
+  if (streq (filename, "-"))
     {
       istream = stdin;
       have_read_stdin = true;
@@ -135,29 +165,26 @@ fold_file (char const *filename, size_t width)
   else
     istream = fopen (filename, "r");
 
-  if (istream == NULL)
+  if (istream == nullptr)
     {
-      error (0, errno, "%s", filename);
+      error (0, errno, "%s", quotef (filename));
       return false;
     }
 
   fadvise (istream, FADVISE_SEQUENTIAL);
+  mbbuf_init (&mbbuf, line_in, sizeof line_in, istream);
 
-  while ((c = getc (istream)) != EOF)
+  mcel_t g;
+  while ((g = mbbuf_get_char (&mbbuf)).ch != MBBUF_EOF)
     {
-      if (offset_out + 1 >= allocated_out)
-        line_out = X2REALLOC (line_out, &allocated_out);
-
-      if (c == '\n')
+      if (g.ch == '\n')
         {
-          line_out[offset_out++] = c;
-          fwrite (line_out, sizeof (char), offset_out, stdout);
+          write_out (line_out, offset_out, /*newline=*/ true);
           column = offset_out = 0;
           continue;
         }
-
     rescan:
-      column = adjust_column (column, c);
+      column = adjust_column (column, g);
 
       if (column > width)
         {
@@ -166,70 +193,83 @@ fold_file (char const *filename, size_t width)
              start the next line. */
           if (break_spaces)
             {
-              bool found_blank = false;
-              size_t logical_end = offset_out;
+              int space_length = 0;
+              idx_t logical_end = offset_out;
+              char *logical_p = line_out;
+              char *logical_lim = logical_p + logical_end;
 
-              /* Look for the last blank. */
-              while (logical_end)
+              for (mcel_t g2; logical_p < logical_lim; logical_p += g2.len)
                 {
-                  --logical_end;
-                  if (isblank (to_uchar (line_out[logical_end])))
+                  g2 = mcel_scan (logical_p, logical_lim);
+                  if (c32isblank (g2.ch) && ! c32isnbspace (g2.ch))
                     {
-                      found_blank = true;
-                      break;
+                      space_length = g2.len;
+                      logical_end = logical_p - line_out;
                     }
                 }
 
-              if (found_blank)
+              if (space_length)
                 {
-                  size_t i;
-
+                  logical_end += space_length;
                   /* Found a blank.  Don't output the part after it. */
-                  logical_end++;
-                  fwrite (line_out, sizeof (char), (size_t) logical_end,
-                          stdout);
-                  putchar ('\n');
+                  write_out (line_out, logical_end, /*newline=*/ true);
                   /* Move the remainder to the beginning of the next line.
                      The areas being copied here might overlap. */
                   memmove (line_out, line_out + logical_end,
                            offset_out - logical_end);
                   offset_out -= logical_end;
-                  for (column = i = 0; i < offset_out; i++)
-                    column = adjust_column (column, line_out[i]);
+                  column = 0;
+                  char *printed_p = line_out;
+                  char *printed_lim = printed_p + offset_out;
+                  for (mcel_t g2; printed_p < printed_lim;
+                       printed_p += g2.len)
+                    {
+                      g2 = mcel_scan (printed_p, printed_lim);
+                      column = adjust_column (column, g2);
+                    }
                   goto rescan;
                 }
             }
 
           if (offset_out == 0)
             {
-              line_out[offset_out++] = c;
+              memcpy (line_out, mbbuf_char_offset (&mbbuf, g), g.len);
+              offset_out += g.len;
               continue;
             }
 
-          line_out[offset_out++] = '\n';
-          fwrite (line_out, sizeof (char), (size_t) offset_out, stdout);
+          write_out (line_out, offset_out, /*newline=*/ true);
           column = offset_out = 0;
           goto rescan;
         }
 
-      line_out[offset_out++] = c;
+      /* This can occur if we have read characters with a width of
+         zero.  */
+      if (sizeof line_out <= offset_out + g.len)
+        {
+          write_out (line_out, offset_out, /*newline=*/ false);
+          offset_out = 0;
+        }
+
+      memcpy (line_out + offset_out, mbbuf_char_offset (&mbbuf, g), g.len);
+      offset_out += g.len;
     }
 
   saved_errno = errno;
+  if (!ferror (istream))
+    saved_errno = 0;
 
   if (offset_out)
-    fwrite (line_out, sizeof (char), (size_t) offset_out, stdout);
+    write_out (line_out, offset_out, /*newline=*/ false);
 
-  if (ferror (istream))
+  if (streq (filename, "-"))
+    clearerr (istream);
+  else if (fclose (istream) != 0 && !saved_errno)
+    saved_errno = errno;
+
+  if (saved_errno)
     {
-      error (0, saved_errno, "%s", filename);
-      if (!STREQ (filename, "-"))
-        fclose (istream);
-      return false;
-    }
-  if (!STREQ (filename, "-") && fclose (istream) == EOF)
-    {
-      error (0, errno, "%s", filename);
+      error (0, saved_errno, "%s", quotef (filename));
       return false;
     }
 
@@ -252,16 +292,20 @@ main (int argc, char **argv)
 
   atexit (close_stdout);
 
-  break_spaces = count_bytes = have_read_stdin = false;
+  break_spaces = have_read_stdin = false;
 
-  while ((optc = getopt_long (argc, argv, shortopts, longopts, NULL)) != -1)
+  while ((optc = getopt_long (argc, argv, shortopts, longopts, nullptr)) != -1)
     {
       char optargbuf[2];
 
       switch (optc)
         {
         case 'b':		/* Count bytes rather than columns. */
-          count_bytes = true;
+          counting_mode = COUNT_BYTES;
+          break;
+
+        case 'c':               /* Count characters rather than columns. */
+          counting_mode = COUNT_CHARACTERS;
           break;
 
         case 's':		/* Break at word boundaries. */
@@ -278,16 +322,11 @@ main (int argc, char **argv)
               optargbuf[1] = '\0';
               optarg = optargbuf;
             }
-          /* Fall through.  */
+          FALLTHROUGH;
         case 'w':		/* Line width. */
-          {
-            unsigned long int tmp_ulong;
-            if (! (xstrtoul (optarg, NULL, 10, &tmp_ulong, "") == LONGINT_OK
-                   && 0 < tmp_ulong && tmp_ulong < SIZE_MAX - TAB_WIDTH))
-              error (EXIT_FAILURE, 0,
-                     _("invalid number of columns: %s"), quote (optarg));
-            width = tmp_ulong;
-          }
+          width = xnumtoumax (optarg, 10, 1, SIZE_MAX - TAB_WIDTH - 1, "",
+                              _("invalid number of columns"), 0,
+                              XTOINT_MIN_RANGE | XTOINT_MAX_RANGE);
           break;
 
         case_GETOPT_HELP_CHAR;
@@ -311,5 +350,5 @@ main (int argc, char **argv)
   if (have_read_stdin && fclose (stdin) == EOF)
     error (EXIT_FAILURE, errno, "-");
 
-  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
+  return ok ? EXIT_SUCCESS : EXIT_FAILURE;
 }
