@@ -1,5 +1,5 @@
 /* pr -- convert text files for printing.
-   Copyright (C) 1988-2013 Free Software Foundation, Inc.
+   Copyright (C) 1988-2025 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,10 +12,10 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /*  By Pete TerMaat, with considerable refinement by Roland Huebner.  */
-
+
 /* Things to watch: Sys V screws up on ...
    pr -n -3 -s: /usr/dict/words
    pr -m -o10 -n /usr/dict/words{,,,}
@@ -46,7 +46,7 @@
       only one POSIX requirement has to be met:
    The default n-separator should be a TAB. The consequence is a
    different width between the number and the text if the output position
-   of the separator changes, i.e. it depends upon the left margin used.
+   of the separator changes, i.e., it depends upon the left margin used.
    That's not nice but easy-to-use together with the defaults of other
    utilities, e.g. sort or cut. - Same as SunOS does.
    -  With multicolumn output
@@ -97,7 +97,7 @@
    adapted to other UNIXes. A violation of downward compatibility has to
    be accepted.
    Some NEW CAPITAL LETTER options ( -J, -S, -W) has been introduced to
-   turn off unexpected interferences of small letter options (-s and -w
+   turn off unexpected interference of small letter options (-s and -w
    together with the three column options).
    -N option and the second argument LAST_PAGE of +FIRST_PAGE offer more
    flexibility; The detailed handling of form feeds set in the input
@@ -143,7 +143,7 @@
                 (a most frequently used form) still work as usual.
 
    -c, --show-control-chars
-                Print unprintable characters as control prefixes.
+                Print nonprintable characters as control prefixes.
                 Control-g is printed as ^G (use hat notation) and
                 octal backslash notation.
 
@@ -272,7 +272,7 @@
                 by form feeds set in the input files.
 
    -v, --show-nonprinting
-                Print unprintable characters as escape sequences. Use
+                Print nonprintable characters as escape sequences. Use
                 octal backslash notation. Control-G becomes \007.
 
    -w PAGE_WIDTH, --width=PAGE_WIDTH
@@ -306,14 +306,14 @@
                 affect -W option.)
 
 */
-
 
 #include <config.h>
 
+#include <ctype.h>
 #include <getopt.h>
 #include <sys/types.h>
 #include "system.h"
-#include "error.h"
+#include "c-ctype.h"
 #include "fadvise.h"
 #include "hard-locale.h"
 #include "mbswidth.h"
@@ -322,6 +322,8 @@
 #include "stdio--.h"
 #include "strftime.h"
 #include "xstrtol.h"
+#include "xstrtol-error.h"
+#include "xdectoint.h"
 
 /* The official name of this program (e.g., no 'g' prefix).  */
 #define PROGRAM_NAME "pr"
@@ -424,6 +426,7 @@ static bool skip_to_page (uintmax_t page);
 static void print_header (void);
 static void pad_across_to (int position);
 static void add_line_number (COLUMN *p);
+static int getoptnum (char const *n_str, int min, char const *errfmt);
 static void getoptarg (char *arg, char switch_char, char *character,
                        int *number);
 static void print_files (int number_of_files, char **av);
@@ -435,13 +438,13 @@ static void init_store_cols (void);
 static void store_columns (void);
 static void balance (int total_stored);
 static void store_char (char c);
-static void pad_down (int lines);
+static void pad_down (unsigned int lines);
 static void read_rest_of_line (COLUMN *p);
 static void skip_read (COLUMN *p, int column_number);
 static void print_char (char c);
 static void cleanup (void);
 static void print_sep_string (void);
-static void separator_string (const char *optarg_S);
+static void separator_string (char const *optarg_S);
 
 /* All of the columns to print.  */
 static COLUMN *column_vector;
@@ -458,7 +461,7 @@ static unsigned int buff_current;
 
 /* The number of characters in buff.
    Used for allocation of buff and to detect overflow of buff. */
-static size_t buff_allocated;
+static idx_t buff_allocated;
 
 /* Array of indices into buff.
    Each entry is an index of the first character of a line.
@@ -466,7 +469,7 @@ static size_t buff_allocated;
    we do column balancing on the last page. */
 static int *line_vector;
 
-/* Array of horizonal positions.
+/* Array of horizontal positions.
    For each line in line_vector, end_vector[line] is the horizontal
    position we are in after printing that line.  We keep track of this
    so that we know how much we need to pad to prepare for the next
@@ -660,11 +663,11 @@ static int number_width;
 /* Buffer sprintf uses to format a line number. */
 static char *number_buff;
 
-/* (-v) True means unprintable characters are printed as escape sequences.
+/* (-v) True means nonprintable characters are printed as escape sequences.
    control-g becomes \007. */
 static bool use_esc_sequence = false;
 
-/* (-c) True means unprintable characters are printed as control prefixes.
+/* (-c) True means nonprintable characters are printed as control prefixes.
    control-g becomes ^G. */
 static bool use_cntrl_prefix = false;
 
@@ -685,7 +688,7 @@ static bool use_col_separator = false;
 /* String used to separate columns if the -S option has been specified.
    Default without -S but together with one of the column options
    -a|COLUMN|-m is a 'space' and with the -J option a 'tab'. */
-static char *col_sep_string = (char *) "";
+static char const *col_sep_string = "";
 static int col_sep_length = 0;
 static char *column_separator = (char *) " ";
 static char *line_separator = (char *) "\t";
@@ -708,6 +711,9 @@ static char *custom_header;
 /* (-D) Date format for the header.  */
 static char const *date_format;
 
+/* The local time zone rules, as per the TZ environment variable.  */
+static timezone_t localtz;
+
 /* Date and file name for the header.  */
 static char *date_text;
 static char const *file_text;
@@ -736,44 +742,51 @@ static char const short_options[] =
 
 static struct option const long_options[] =
 {
-  {"pages", required_argument, NULL, PAGES_OPTION},
-  {"columns", required_argument, NULL, COLUMNS_OPTION},
-  {"across", no_argument, NULL, 'a'},
-  {"show-control-chars", no_argument, NULL, 'c'},
-  {"double-space", no_argument, NULL, 'd'},
-  {"date-format", required_argument, NULL, 'D'},
-  {"expand-tabs", optional_argument, NULL, 'e'},
-  {"form-feed", no_argument, NULL, 'f'},
-  {"header", required_argument, NULL, 'h'},
-  {"output-tabs", optional_argument, NULL, 'i'},
-  {"join-lines", no_argument, NULL, 'J'},
-  {"length", required_argument, NULL, 'l'},
-  {"merge", no_argument, NULL, 'm'},
-  {"number-lines", optional_argument, NULL, 'n'},
-  {"first-line-number", required_argument, NULL, 'N'},
-  {"indent", required_argument, NULL, 'o'},
-  {"no-file-warnings", no_argument, NULL, 'r'},
-  {"separator", optional_argument, NULL, 's'},
-  {"sep-string", optional_argument, NULL, 'S'},
-  {"omit-header", no_argument, NULL, 't'},
-  {"omit-pagination", no_argument, NULL, 'T'},
-  {"show-nonprinting", no_argument, NULL, 'v'},
-  {"width", required_argument, NULL, 'w'},
-  {"page-width", required_argument, NULL, 'W'},
+  {"pages", required_argument, nullptr, PAGES_OPTION},
+  {"columns", required_argument, nullptr, COLUMNS_OPTION},
+  {"across", no_argument, nullptr, 'a'},
+  {"show-control-chars", no_argument, nullptr, 'c'},
+  {"double-space", no_argument, nullptr, 'd'},
+  {"date-format", required_argument, nullptr, 'D'},
+  {"expand-tabs", optional_argument, nullptr, 'e'},
+  {"form-feed", no_argument, nullptr, 'f'},
+  {"header", required_argument, nullptr, 'h'},
+  {"output-tabs", optional_argument, nullptr, 'i'},
+  {"join-lines", no_argument, nullptr, 'J'},
+  {"length", required_argument, nullptr, 'l'},
+  {"merge", no_argument, nullptr, 'm'},
+  {"number-lines", optional_argument, nullptr, 'n'},
+  {"first-line-number", required_argument, nullptr, 'N'},
+  {"indent", required_argument, nullptr, 'o'},
+  {"no-file-warnings", no_argument, nullptr, 'r'},
+  {"separator", optional_argument, nullptr, 's'},
+  {"sep-string", optional_argument, nullptr, 'S'},
+  {"omit-header", no_argument, nullptr, 't'},
+  {"omit-pagination", no_argument, nullptr, 'T'},
+  {"show-nonprinting", no_argument, nullptr, 'v'},
+  {"width", required_argument, nullptr, 'w'},
+  {"page-width", required_argument, nullptr, 'W'},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
-  {NULL, 0, NULL, 0}
+  {nullptr, 0, nullptr, 0}
 };
 
+static _Noreturn void
+integer_overflow (void)
+{
+  error (EXIT_FAILURE, 0, _("integer overflow"));
+}
+
 /* Return the number of columns that have either an open file or
    stored lines. */
 
-static int _GL_ATTRIBUTE_PURE
+ATTRIBUTE_PURE
+static unsigned int
 cols_ready_to_print (void)
 {
   COLUMN *q;
-  int i;
-  int n;
+  unsigned int i;
+  unsigned int n;
 
   n = 0;
   for (q = column_vector, i = 0; i < columns; ++q, ++i)
@@ -820,44 +833,40 @@ first_last_page (int oi, char c, char const *pages)
 
 /* Parse column count string S, and if it's valid (1 or larger and
    within range of the type of 'columns') set the global variables
-   columns and explicit_columns and return true.
-   Otherwise, exit with a diagnostic.  */
+   columns and explicit_columns.  Otherwise, exit with a diagnostic.  */
+
 static void
 parse_column_count (char const *s)
 {
-  long int tmp_long;
-  if (xstrtol (s, NULL, 10, &tmp_long, "") != LONGINT_OK
-      || !(1 <= tmp_long && tmp_long <= INT_MAX))
-    error (EXIT_FAILURE, 0,
-           _("invalid number of columns: %s"), quote (s));
-
-  columns = tmp_long;
+  columns = getoptnum (s, 1, _("invalid number of columns"));
   explicit_columns = true;
 }
 
 /* Estimate length of col_sep_string with option -S.  */
 
 static void
-separator_string (const char *optarg_S)
+separator_string (char const *optarg_S)
 {
-  col_sep_length = (int) strlen (optarg_S);
-  col_sep_string = xmalloc (col_sep_length + 1);
-  strcpy (col_sep_string, optarg_S);
+  size_t len = strlen (optarg_S);
+  if (INT_MAX < len)
+    integer_overflow ();
+  col_sep_length = len;
+  col_sep_string = optarg_S;
 }
 
 int
 main (int argc, char **argv)
 {
-  int n_files;
+  unsigned int n_files;
   bool old_options = false;
   bool old_w = false;
   bool old_s = false;
   char **file_names;
 
   /* Accumulate the digits of old-style options like -99.  */
-  char *column_count_string = NULL;
-  size_t n_digits = 0;
-  size_t n_alloc = 0;
+  char *column_count_string = nullptr;
+  idx_t n_digits = 0;
+  idx_t n_alloc = 0;
 
   initialize_main (&argc, &argv);
   set_program_name (argv[0]);
@@ -869,8 +878,8 @@ main (int argc, char **argv)
 
   n_files = 0;
   file_names = (argc > 1
-                ? xmalloc ((argc - 1) * sizeof (char *))
-                : NULL);
+                ? xnmalloc (argc - 1, sizeof (char *))
+                : nullptr);
 
   while (true)
     {
@@ -879,12 +888,12 @@ main (int argc, char **argv)
       if (c == -1)
         break;
 
-      if (ISDIGIT (c))
+      if (c_isdigit (c))
         {
           /* Accumulate column-count digits specified via old-style options. */
           if (n_digits + 1 >= n_alloc)
-            column_count_string
-              = X2REALLOC (column_count_string, &n_alloc);
+            column_count_string = xpalloc (column_count_string, &n_alloc, 2, -1,
+                                           sizeof *column_count_string);
           column_count_string[n_digits++] = c;
           column_count_string[n_digits] = '\0';
           continue;
@@ -920,7 +929,7 @@ main (int argc, char **argv)
                short-named option syntax, e.g., -9, ensure that this
                long-name-specified value overrides it.  */
             free (column_count_string);
-            column_count_string = NULL;
+            column_count_string = nullptr;
             n_alloc = 0;
             break;
           }
@@ -966,18 +975,10 @@ main (int argc, char **argv)
           join_lines = true;
           break;
         case 'l':
-          {
-            long int tmp_long;
-            if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
-                || tmp_long <= 0 || tmp_long > INT_MAX)
-              {
-                error (EXIT_FAILURE, 0,
-                       _("'-l PAGE_LENGTH' invalid number of lines: %s"),
-                       quote (optarg));
-              }
-            lines_per_page = tmp_long;
-            break;
-          }
+          lines_per_page
+            = getoptnum (optarg, 1,
+                         _("'-l PAGE_LENGTH' invalid number of lines"));
+          break;
         case 'm':
           parallel_files = true;
           storing_columns = false;
@@ -990,28 +991,14 @@ main (int argc, char **argv)
           break;
         case 'N':
           skip_count = false;
-          {
-            long int tmp_long;
-            if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
-                || tmp_long > INT_MAX)
-              {
-                error (EXIT_FAILURE, 0,
-                       _("'-N NUMBER' invalid starting line number: %s"),
-                       quote (optarg));
-              }
-            start_line_num = tmp_long;
-            break;
-          }
+          start_line_num
+            = getoptnum (optarg, INT_MIN,
+                         _("'-N NUMBER' invalid starting line number"));
+          break;
         case 'o':
-          {
-            long int tmp_long;
-            if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
-                || tmp_long < 0 || tmp_long > INT_MAX)
-              error (EXIT_FAILURE, 0,
-                     _("'-o MARGIN' invalid line offset: %s"), quote (optarg));
-            chars_per_margin = tmp_long;
-            break;
-          }
+          chars_per_margin = getoptnum (optarg, 0,
+                                        _("'-o MARGIN' invalid line offset"));
+          break;
         case 'r':
           ignore_failed_opens = true;
           break;
@@ -1024,7 +1011,7 @@ main (int argc, char **argv)
         case 'S':
           old_s = false;
           /* Reset an additional input of -s, -S dominates -s */
-          col_sep_string = bad_cast ("");
+          col_sep_string = "";
           col_sep_length = 0;
           use_col_separator = true;
           if (optarg)
@@ -1045,29 +1032,20 @@ main (int argc, char **argv)
           old_options = true;
           old_w = true;
           {
-            long int tmp_long;
-            if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
-                || tmp_long <= 0 || tmp_long > INT_MAX)
-              error (EXIT_FAILURE, 0,
-                     _("'-w PAGE_WIDTH' invalid number of characters: %s"),
-                     quote (optarg));
-            if (!truncate_lines)
-              chars_per_line = tmp_long;
-            break;
+            int tmp_cpl
+              = getoptnum (optarg, 1,
+                           _("'-w PAGE_WIDTH' invalid number of characters"));
+            if (! truncate_lines)
+              chars_per_line = tmp_cpl;
           }
+          break;
         case 'W':
           old_w = false;			/* dominates -w */
           truncate_lines = true;
-          {
-            long int tmp_long;
-            if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
-                || tmp_long <= 0 || tmp_long > INT_MAX)
-              error (EXIT_FAILURE, 0,
-                     _("'-W PAGE_WIDTH' invalid number of characters: %s"),
-                     quote (optarg));
-            chars_per_line = tmp_long;
-            break;
-          }
+          chars_per_line
+            = getoptnum (optarg, 1,
+                         _("'-W PAGE_WIDTH' invalid number of characters"));
+          break;
         case_GETOPT_HELP_CHAR;
         case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
         default:
@@ -1087,17 +1065,19 @@ main (int argc, char **argv)
                    ? "%b %e %H:%M %Y"
                    : "%Y-%m-%d %H:%M");
 
+  localtz = tzalloc (getenv ("TZ"));
+
   /* Now we can set a reasonable initial value: */
   if (first_page_number == 0)
     first_page_number = 1;
 
   if (parallel_files && explicit_columns)
     error (EXIT_FAILURE, 0,
-         _("cannot specify number of columns when printing in parallel"));
+           _("cannot specify number of columns when printing in parallel"));
 
   if (parallel_files && print_across_flag)
     error (EXIT_FAILURE, 0,
-       _("cannot specify both printing across and printing in parallel"));
+           _("cannot specify both printing across and printing in parallel"));
 
 /* Translate some old short options to new/long options.
    To meet downward compatibility with other UNIX pr utilities
@@ -1152,7 +1132,7 @@ main (int argc, char **argv)
   if (n_files == 0)
     {
       /* No file arguments specified;  read from standard input.  */
-      print_files (0, NULL);
+      print_files (0, nullptr);
     }
   else
     {
@@ -1160,8 +1140,7 @@ main (int argc, char **argv)
         print_files (n_files, file_names);
       else
         {
-          unsigned int i;
-          for (i = 0; i < n_files; i++)
+          for (unsigned int i = 0; i < n_files; i++)
             print_files (1, &file_names[i]);
         }
     }
@@ -1170,9 +1149,16 @@ main (int argc, char **argv)
 
   if (have_read_stdin && fclose (stdin) == EOF)
     error (EXIT_FAILURE, errno, _("standard input"));
-  if (failed_opens)
-    exit (EXIT_FAILURE);
-  exit (EXIT_SUCCESS);
+  main_exit (failed_opens ? EXIT_FAILURE : EXIT_SUCCESS);
+}
+
+/* Parse numeric arguments, ensuring MIN <= number <= INT_MAX.  */
+
+static int
+getoptnum (char const *n_str, int min, char const *err)
+{
+  return xnumtoimax (n_str, 10, min, INT_MAX, "", err, 0,
+                     min <= 0 ? 0 : XTOINT_MIN_RANGE);
 }
 
 /* Parse options of the form -scNNN.
@@ -1184,15 +1170,28 @@ main (int argc, char **argv)
 static void
 getoptarg (char *arg, char switch_char, char *character, int *number)
 {
-  if (!ISDIGIT (*arg))
+  if (!*arg)
+    {
+      error (0, 0, _("'-%c': Invalid argument: %s"), switch_char, quote (arg));
+      usage (EXIT_FAILURE);
+    }
+
+  if (!c_isdigit (*arg))
     *character = *arg++;
   if (*arg)
     {
       long int tmp_long;
-      if (xstrtol (arg, NULL, 10, &tmp_long, "") != LONGINT_OK
-          || tmp_long <= 0 || tmp_long > INT_MAX)
+      strtol_error e = xstrtol (arg, nullptr, 10, &tmp_long, "");
+      if (e == LONGINT_OK)
         {
-          error (0, 0,
+          if (tmp_long <= 0)
+            e = LONGINT_INVALID;
+          else if (INT_MAX < tmp_long)
+            e = LONGINT_OVERFLOW;
+        }
+      if (e != LONGINT_OK)
+        {
+          error (0, e & LONGINT_OVERFLOW ? EOVERFLOW : 0,
              _("'-%c' extra characters or invalid number in the argument: %s"),
                  switch_char, quote (arg));
           usage (EXIT_FAILURE);
@@ -1200,7 +1199,7 @@ getoptarg (char *arg, char switch_char, char *character, int *number)
       *number = tmp_long;
     }
 }
-
+
 /* Set parameters related to formatting. */
 
 static void
@@ -1218,7 +1217,7 @@ init_parameters (int number_of_files)
     lines_per_body = lines_per_page;
 
   if (double_space)
-    lines_per_body = lines_per_body / 2;
+    lines_per_body = MAX (1, lines_per_body / 2);
 
   /* If input is stdin, cannot print parallel files.  BSD dumps core
      on this. */
@@ -1249,10 +1248,12 @@ init_parameters (int number_of_files)
         }
       /* It's rather pointless to define a TAB separator with column
          alignment */
-      else if (!join_lines && *col_sep_string == '\t')
+      else if (!join_lines && col_sep_length == 1 && *col_sep_string == '\t')
         col_sep_string = column_separator;
 
       truncate_lines = true;
+      if (! (col_sep_length == 1 && *col_sep_string == '\t'))
+        untabify_input = true;
       tabify_output = true;
     }
   else
@@ -1286,8 +1287,13 @@ init_parameters (int number_of_files)
         chars_used_by_number = number_width;
     }
 
-  chars_per_column = (chars_per_line - chars_used_by_number
-                      - (columns - 1) * col_sep_length) / columns;
+  int sep_chars, useful_chars;
+  if (ckd_mul (&sep_chars, columns - 1, col_sep_length))
+    sep_chars = INT_MAX;
+  if (ckd_sub (&useful_chars, chars_per_line - chars_used_by_number,
+               sep_chars))
+    useful_chars = 0;
+  chars_per_column = useful_chars / columns;
 
   if (chars_per_column < 1)
     error (EXIT_FAILURE, 0, _("page width too narrow"));
@@ -1307,7 +1313,7 @@ init_parameters (int number_of_files)
   free (clump_buff);
   clump_buff = xmalloc (MAX (8, chars_per_input_tab));
 }
-
+
 /* Open the necessary files,
    maintaining a COLUMN structure for each column.
 
@@ -1322,10 +1328,7 @@ init_parameters (int number_of_files)
 static bool
 init_fps (int number_of_files, char **av)
 {
-  int i, files_left;
   COLUMN *p;
-  FILE *firstfp;
-  char const *firstname;
 
   total_files = 0;
 
@@ -1334,7 +1337,7 @@ init_fps (int number_of_files, char **av)
 
   if (parallel_files)
     {
-      files_left = number_of_files;
+      int files_left = number_of_files;
       for (p = column_vector; files_left--; ++p, ++av)
         {
           if (! open_file (*av, p))
@@ -1369,8 +1372,9 @@ init_fps (int number_of_files, char **av)
           p->lines_stored = 0;
         }
 
-      firstname = p->name;
-      firstfp = p->fp;
+      char const *firstname = p->name;
+      FILE *firstfp = p->fp;
+      int i;
       for (i = columns - 1, ++p; i; --i, ++p)
         {
           p->name = firstname;
@@ -1383,7 +1387,7 @@ init_fps (int number_of_files, char **av)
   files_ready_to_read = total_files;
   return true;
 }
-
+
 /* Determine print_func and char_func, the functions
    used by each column for printing and/or storing.
 
@@ -1470,7 +1474,7 @@ init_funcs (void)
   p->numbered = numbered_lines && (!parallel_files || i == 1);
   p->start_position = h;
 }
-
+
 /* Open a file.  Return true if successful.
 
    With each file p, p->full_page_printed is initialized,
@@ -1479,7 +1483,7 @@ init_funcs (void)
 static bool
 open_file (char *name, COLUMN *p)
 {
-  if (STREQ (name, "-"))
+  if (streq (name, "-"))
     {
       p->name = _("standard input");
       p->fp = stdin;
@@ -1490,11 +1494,11 @@ open_file (char *name, COLUMN *p)
       p->name = name;
       p->fp = fopen (name, "r");
     }
-  if (p->fp == NULL)
+  if (p->fp == nullptr)
     {
       failed_opens = true;
       if (!ignore_failed_opens)
-        error (0, errno, "%s", name);
+        error (0, errno, "%s", quotef (name));
       return false;
     }
   fadvise (p->fp, FADVISE_SEQUENTIAL);
@@ -1517,10 +1521,16 @@ close_file (COLUMN *p)
 
   if (p->status == CLOSED)
     return;
-  if (ferror (p->fp))
-    error (EXIT_FAILURE, errno, "%s", p->name);
-  if (fileno (p->fp) != STDIN_FILENO && fclose (p->fp) != 0)
-    error (EXIT_FAILURE, errno, "%s", p->name);
+
+  int err = errno;
+  if (!ferror (p->fp))
+    err = 0;
+  if (fileno (p->fp) == STDIN_FILENO)
+    clearerr (p->fp);
+  else if (fclose (p->fp) != 0 && !err)
+    err = errno;
+  if (err)
+    error (EXIT_FAILURE, err, "%s", quotef (p->name));
 
   if (!parallel_files)
     {
@@ -1594,7 +1604,7 @@ reset_status (void)
         files_ready_to_read = 1;
     }
 }
-
+
 /* Print a single file, or multiple files in parallel.
 
    Set up the list of columns, opening the necessary files.
@@ -1629,7 +1639,7 @@ print_files (int number_of_files, char **av)
   while (print_page ())
     ;
 }
-
+
 /* Initialize header information.
    If DESC is non-negative, it is a file descriptor open to
    FILENAME for reading.  */
@@ -1637,14 +1647,14 @@ print_files (int number_of_files, char **av)
 static void
 init_header (char const *filename, int desc)
 {
-  char *buf = NULL;
+  char *buf = nullptr;
   struct stat st;
   struct timespec t;
   int ns;
-  struct tm *tm;
+  struct tm tm;
 
   /* If parallel files or standard input, use current date. */
-  if (STREQ (filename, "-"))
+  if (streq (filename, "-"))
     desc = -1;
   if (0 <= desc && fstat (desc, &st) == 0)
     t = get_stat_mtime (&st);
@@ -1657,18 +1667,21 @@ init_header (char const *filename, int desc)
     }
 
   ns = t.tv_nsec;
-  tm = localtime (&t.tv_sec);
-  if (tm == NULL)
+  if (localtime_rz (localtz, &t.tv_sec, &tm))
     {
-      buf = xmalloc (INT_BUFSIZE_BOUND (long int)
-                     + MAX (10, INT_BUFSIZE_BOUND (int)));
-      sprintf (buf, "%ld.%09d", (long int) t.tv_sec, ns);
+      ptrdiff_t len = nstrftime (nullptr, MIN (PTRDIFF_MAX, SIZE_MAX),
+                                 date_format, &tm, localtz, ns);
+      if (0 <= len)
+        {
+          buf = ximalloc (len + 1);
+          nstrftime (buf, len + 1, date_format, &tm, localtz, ns);
+        }
     }
-  else
+  if (!buf)
     {
-      size_t bufsize = nstrftime (NULL, SIZE_MAX, date_format, tm, 0, ns) + 1;
-      buf = xmalloc (bufsize);
-      nstrftime (buf, bufsize, date_format, tm, 0, ns);
+      char secbuf[INT_BUFSIZE_BOUND (intmax_t)];
+      buf = xmalloc (sizeof secbuf + MAX (10, INT_BUFSIZE_BOUND (int)));
+      sprintf (buf, "%s.%09d", timetostr (t.tv_sec, secbuf), ns);
     }
 
   free (date_text);
@@ -1678,7 +1691,7 @@ init_header (char const *filename, int desc)
                             - mbswidth (date_text, 0)
                             - mbswidth (file_text, 0));
 }
-
+
 /* Set things up for printing a page
 
    Scan through the columns ...
@@ -1738,7 +1751,7 @@ static void
 align_column (COLUMN *p)
 {
   padding_not_printed = p->start_position;
-  if (padding_not_printed - col_sep_length > 0)
+  if (col_sep_length < padding_not_printed)
     {
       pad_across_to (padding_not_printed - col_sep_length);
       padding_not_printed = ANYWHERE;
@@ -1817,7 +1830,7 @@ print_page (void)
               --p->lines_to_print;
               if (p->lines_to_print <= 0)
                 {
-                  if (cols_ready_to_print () <= 0)
+                  if (cols_ready_to_print () == 0)
                     break;
                 }
 
@@ -1851,7 +1864,7 @@ print_page (void)
           --lines_left_on_page;
         }
 
-      if (cols_ready_to_print () <= 0 && !extremities)
+      if (cols_ready_to_print () == 0 && !extremities)
         break;
 
       if (double_space && pv)
@@ -1876,6 +1889,9 @@ print_page (void)
       print_a_FF = false;
     }
 
+  if (ferror (stdout))
+    write_error ();
+
   if (last_page_number < ++page_number)
     return false;		/* Stop printing with LAST_PAGE */
 
@@ -1883,7 +1899,7 @@ print_page (void)
 
   return true;			/* More pages to go. */
 }
-
+
 /* Allocate space for storing columns.
 
    This is necessary when printing multiple columns from a single file.
@@ -1901,21 +1917,24 @@ print_page (void)
 static void
 init_store_cols (void)
 {
-  int total_lines = lines_per_body * columns;
-  int chars_if_truncate = total_lines * (chars_per_column + 1);
+  /* Tune this.  */
+  int total_lines, total_lines_1, chars_per_column_1, chars_if_truncate;
+  if (ckd_mul (&total_lines, lines_per_body, columns)
+      || ckd_add (&total_lines_1, total_lines, 1)
+      || ckd_add (&chars_per_column_1, chars_per_column, 1)
+      || ckd_mul (&chars_if_truncate, total_lines, chars_per_column_1)
+      || ckd_mul (&buff_allocated, chars_if_truncate, use_col_separator + 1))
+    integer_overflow ();
 
   free (line_vector);
   /* FIXME: here's where it was allocated.  */
-  line_vector = xmalloc ((total_lines + 1) * sizeof *line_vector);
+  line_vector = xnmalloc (total_lines_1, sizeof *line_vector);
 
   free (end_vector);
-  end_vector = xmalloc (total_lines * sizeof *end_vector);
+  end_vector = xnmalloc (total_lines, sizeof *end_vector);
 
   free (buff);
-  buff_allocated = (use_col_separator
-                    ? 2 * chars_if_truncate
-                    : chars_if_truncate);	/* Tune this. */
-  buff = xmalloc (buff_allocated);
+  buff = ximalloc (buff_allocated);
 }
 
 /* Store all but the rightmost column.
@@ -2009,7 +2028,7 @@ store_char (char c)
   if (buff_current >= buff_allocated)
     {
       /* May be too generous. */
-      buff = X2REALLOC (buff, &buff_allocated);
+      buff = xpalloc (buff, &buff_allocated, 1, -1, sizeof *buff);
     }
   buff[buff_current++] = c;
 }
@@ -2057,7 +2076,7 @@ add_line_number (COLUMN *p)
   if (truncate_lines && !parallel_files)
     input_position += number_width;
 }
-
+
 /* Print (or store) padding until the current horizontal position
    is position. */
 
@@ -2082,14 +2101,12 @@ pad_across_to (int position)
    Otherwise, use newlines. */
 
 static void
-pad_down (int lines)
+pad_down (unsigned int lines)
 {
-  int i;
-
   if (use_form_feed)
     putchar ('\f');
   else
-    for (i = lines; i; --i)
+    for (unsigned int i = lines; i; --i)
       putchar ('\n');
 }
 
@@ -2193,7 +2210,7 @@ skip_read (COLUMN *p, int column_number)
     if ((!parallel_files || column_number == 1) && !single_ff)
       ++line_count;
 }
-
+
 /* If we're tabifying output,
 
    When print_char encounters white space it keeps track
@@ -2228,11 +2245,9 @@ print_white_space (void)
 static void
 print_sep_string (void)
 {
-  char *s;
+  char const *s = col_sep_string;
   int l = col_sep_length;
 
-  s = col_sep_string;
-
   if (separators_not_printed <= 0)
     {
       /* We'll be starting a line with chars_per_margin, anything else? */
@@ -2277,6 +2292,9 @@ print_clump (COLUMN *p, int n, char *clump)
 {
   while (n--)
     (p->char_func) (*clump++);
+
+  if (ferror (stdout))
+    write_error ();
 }
 
 /* Print a character.
@@ -2319,14 +2337,12 @@ print_char (char c)
 static bool
 skip_to_page (uintmax_t page)
 {
-  uintmax_t n;
-  int i;
-  int j;
-  COLUMN *p;
-
-  for (n = 1; n < page; ++n)
+  for (uintmax_t n = 1; n < page; ++n)
     {
-      for (i = 1; i < lines_per_body; ++i)
+      COLUMN *p;
+      int j;
+
+      for (int i = 1; i < lines_per_body; ++i)
         {
           for (j = 1, p = column_vector; j <= columns; ++j, ++p)
             if (p->status == OPEN)
@@ -2350,8 +2366,7 @@ skip_to_page (uintmax_t page)
           /* It's very helpful, normally the total number of pages is
              not known in advance.  */
           error (0, 0,
-                 _("starting page number %"PRIuMAX
-                   " exceeds page count %"PRIuMAX),
+                 _("starting page number %ju exceeds page count %ju"),
                  page, n);
           break;
         }
@@ -2380,9 +2395,9 @@ print_header (void)
     error (EXIT_FAILURE, 0, _("page number overflow"));
 
   /* The translator must ensure that formatting the translation of
-     "Page %"PRIuMAX does not generate more than (sizeof page_text - 1)
+     "Page %ju" does not generate more than (sizeof page_text - 1)
      bytes.  */
-  sprintf (page_text, _("Page %"PRIuMAX), page_number);
+  sprintf (page_text, _("Page %ju"), page_number);
   available_width = header_width_available - mbswidth (page_text, 0);
   available_width = MAX (0, available_width);
   lhs_spaces = available_width >> 1;
@@ -2422,7 +2437,7 @@ static bool
 read_line (COLUMN *p)
 {
   int c;
-  int chars IF_LINT ( = 0);
+  int chars IF_LINT (= 0);
   int last_input_position;
   int j, k;
   COLUMN *q;
@@ -2492,7 +2507,7 @@ read_line (COLUMN *p)
           align_empty_cols = false;
         }
 
-      if (padding_not_printed - col_sep_length > 0)
+      if (col_sep_length < padding_not_printed)
         {
           pad_across_to (padding_not_printed - col_sep_length);
           padding_not_printed = ANYWHERE;
@@ -2561,7 +2576,6 @@ static bool
 print_stored (COLUMN *p)
 {
   COLUMN *q;
-  int i;
 
   int line = p->current_line++;
   char *first = &buff[line_vector[line]];
@@ -2585,6 +2599,7 @@ print_stored (COLUMN *p)
 
   if (p->status == FF_FOUND)
     {
+      int i;
       for (i = 1, q = column_vector; i <= columns; ++i, ++q)
         q->status = ON_HOLD;
       if (column_vector->lines_to_print <= 0)
@@ -2595,7 +2610,7 @@ print_stored (COLUMN *p)
         }
     }
 
-  if (padding_not_printed - col_sep_length > 0)
+  if (col_sep_length < padding_not_printed)
     {
       pad_across_to (padding_not_printed - col_sep_length);
       padding_not_printed = ANYWHERE;
@@ -2626,7 +2641,7 @@ print_stored (COLUMN *p)
    sequences or control prefixes.
 
    Note: the width of a clump is not necessarily equal to the number of
-   characters in clump_buff.  (e.g, the width of '\b' is -1, while the
+   characters in clump_buff.  (e.g., the width of '\b' is -1, while the
    number of characters is 1.) */
 
 static int
@@ -2739,7 +2754,7 @@ cleanup (void)
   free (end_vector);
   free (buff);
 }
-
+
 /* Complain, print a usage message, and die. */
 
 void
@@ -2758,6 +2773,7 @@ Usage: %s [OPTION]... [FILE]...\n\
 Paginate or columnate FILE(s) for printing.\n\
 "), stdout);
 
+      emit_stdin_note ();
       emit_mandatory_arg_note ();
 
       fputs (_("\
@@ -2798,7 +2814,10 @@ Paginate or columnate FILE(s) for printing.\n\
       fputs (_("\
   -l, --length=PAGE_LENGTH\n\
                     set the page length to PAGE_LENGTH (66) lines\n\
-                    (default number of lines of text 56, and with -F 63)\n\
+                    (default number of lines of text 56, and with -F 63).\n\
+                    implies -t if PAGE_LENGTH <= 10\n\
+"), stdout);
+      fputs (_("\
   -m, --merge       print all files in parallel, one in each column,\n\
                     truncate lines, but join lines of full length with -J\n\
 "), stdout);
@@ -2820,7 +2839,8 @@ Paginate or columnate FILE(s) for printing.\n\
       fputs (_("\
   -s[CHAR], --separator[=CHAR]\n\
                     separate columns by a single character, default for CHAR\n\
-                    is the <TAB> character without -w and \'no char\' with -w\n\
+                    is the <TAB> character without -w and \'no char\' with -w.\
+\n\
                     -s[CHAR] turns off line truncation of all 3 column\n\
                     options (-COLUMN|-a -COLUMN|-m) except -w is set\n\
 "), stdout);
@@ -2829,7 +2849,10 @@ Paginate or columnate FILE(s) for printing.\n\
                     separate columns by STRING,\n\
                     without -S: Default separator <TAB> with -J and <space>\n\
                     otherwise (same as -S\" \"), no effect on column options\n\
-  -t, --omit-header  omit page headers and trailers\n\
+"), stdout);
+      fputs (_("\
+  -t, --omit-header  omit page headers and trailers;\n\
+                     implied if PAGE_LENGTH <= 10\n\
 "), stdout);
       fputs (_("\
   -T, --omit-pagination\n\
@@ -2849,12 +2872,7 @@ Paginate or columnate FILE(s) for printing.\n\
 "), stdout);
       fputs (HELP_OPTION_DESCRIPTION, stdout);
       fputs (VERSION_OPTION_DESCRIPTION, stdout);
-      fputs (_("\
-\n\
--t is implied if PAGE_LENGTH <= 10.  With no FILE, or when FILE is -, read\n\
-standard input.\n\
-"), stdout);
-      emit_ancillary_info ();
+      emit_ancillary_info (PROGRAM_NAME);
     }
   exit (status);
 }
