@@ -3,7 +3,7 @@
 # Show that du fails with ELOOP (Too many levels of symbolic links)
 # when it encounters that condition.
 
-# Copyright (C) 2006-2013 Free Software Foundation, Inc.
+# Copyright (C) 2006-2025 Free Software Foundation, Inc.
 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
 # GNU General Public License for more details.
 
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 . "${srcdir=.}/tests/init.sh"; path_prepend_ ./src
 print_ver_ du
@@ -42,7 +42,7 @@ for i in $dir_list $(expr $n + 1); do
   esac
   i_minus_1=$i
 done
-echo foo > $i
+echo foo > $i || framework_failure_
 
 # If a system can handle this many symlinks in a file name,
 # just skip this test.
@@ -63,9 +63,10 @@ too_many=$(sed 's/.*: //' err)
 # With coreutils-5.94 we get the desired diagnostic:
 # du: cannot access '1/s/s/s/.../s': Too many levels of symbolic links
 du -L 1 > /dev/null 2> out1 && fail=1
-sed "s, .1/s/s/s/[/s]*',," out1 > out || fail=1
+sed "s, .1/s/s/s/[/s]*',," out1 > out2 || framework_failure_
+sed "s/cannot read directory/cannot access/" out2 > out || framework_failure_
 
-echo "du: cannot access: $too_many" > exp || fail=1
+echo "du: cannot access: $too_many" > exp || framework_failure_
 
 compare exp out || fail=1
 
